# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import mock

from oslo_utils import uuidutils

from neutron.tests.unit.api.v2 import test_base
from neutron.tests.unit.extensions import base as test_extensions_base
from neutron_lib.api.definitions import bgpvpn as bgpvpn_def
from webob import exc

from networking_bgpvpn.neutron.extensions import bgpvpn

_uuid = uuidutils.generate_uuid
_get_path = test_base._get_path
BGPVPN_PREFIX = 'bgpvpn'
BGPVPN_URI = BGPVPN_PREFIX + '/' + 'bgpvpns'
BGPVPN_PLUGIN_BASE_NAME = (
    bgpvpn.BGPVPNPluginBase.__module__ + '.' +
    bgpvpn.BGPVPNPluginBase.__name__)


class BgpvpnExtensionTestCase(test_extensions_base.ExtensionTestCase):
    fmt = 'json'

    def setUp(self):
        super(BgpvpnExtensionTestCase, self).setUp()
        plural_mappings = {'bgpvpn': 'bgpvpns'}
        self._setUpExtension(
            BGPVPN_PLUGIN_BASE_NAME,
            bgpvpn_def.LABEL,
            bgpvpn_def.RESOURCE_ATTRIBUTE_MAP,
            bgpvpn.Bgpvpn,
            BGPVPN_PREFIX,
            plural_mappings=plural_mappings,
            translate_resource_name=True)
        self.instance = self.plugin.return_value
        self.bgpvpn_id = _uuid()
        self.net_id = _uuid()
        self.router_id = _uuid()
        self.net_assoc_id = _uuid()
        self.router_assoc_id = _uuid()
        self.NET_ASSOC_URI = BGPVPN_URI + '/' + self.bgpvpn_id + \
            '/network_associations'
        self.ROUTER_ASSOC_URI = BGPVPN_URI + '/' + self.bgpvpn_id + \
            '/router_associations'

    def test_bgpvpn_create(self):
        bgpvpn_id = _uuid()
        data = {
            'bgpvpn': {'name': 'bgpvpn1',
                       'type': 'l3',
                       'route_targets': ['1234:56'],
                       'tenant_id': _uuid()}
        }
        expected_ret_val = copy.copy(data['bgpvpn'])
        expected_ret_val['import_targets'] = []
        expected_ret_val['export_targets'] = []
        expected_ret_val['route_distinguishers'] = []
        expected_call_args = copy.copy(expected_ret_val)
        expected_ret_val.update({'id': bgpvpn_id})

        self.instance.create_bgpvpn.return_value = expected_ret_val
        res = self.api.post(_get_path(BGPVPN_URI, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        self.assertTrue(self.instance.create_bgpvpn.called)
        self.assertDictSupersetOf(
            expected_call_args,
            self.instance.create_bgpvpn.call_args[1]['bgpvpn']['bgpvpn'])
        self.assertEqual(res.status_int, exc.HTTPCreated.code)
        res = self.deserialize(res)
        self.assertIn('bgpvpn', res)
        self.assertDictSupersetOf(expected_ret_val, res['bgpvpn'])

    def test_bgpvpn_create_with_malformatted_route_target(self):
        data = {
            'bgpvpn': {'name': 'bgpvpn1',
                       'type': 'l3',
                       'route_targets': ['ASN:NN'],
                       'tenant_id': _uuid()}
        }

        res = self.api.post(_get_path(BGPVPN_URI, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt,
                            expect_errors=True)

        self.assertEqual(res.status_int, exc.HTTPBadRequest.code)

    def _data_for_invalid_rtdt(self, field):
        values = [[':1'],
                  ['1:'],
                  ['42'],
                  ['65536:123456'],
                  ['123.456.789.123:65535'],
                  ['4294967296:65535'],
                  ['1.1.1.1:655351'],
                  ['4294967295:65536'],
                  ]
        for value in values:
            yield {'bgpvpn': {'name': 'bgpvpn1',
                              'type': 'l3',
                              field: value,
                              'tenant_id': _uuid()}
                   }

    def _data_for_valid_rtdt(self, field):
        values = [['1:1'],
                  ['1:4294967295'],
                  ['65535:0'],
                  ['65535:4294967295'],
                  ['1.1.1.1:1'],
                  ['1.1.1.1:65535'],
                  ['4294967295:0'],
                  ['65536:65535'],
                  ['4294967295:65535'],
                  ]
        for value in values:
            yield {'bgpvpn': {'name': 'bgpvpn1',
                              'type': 'l3',
                              field: value,
                              'tenant_id': _uuid()}
                   }

    def _test_invalid_field(self, field):
        for data in self._data_for_invalid_rtdt(field):
            res = self.api.post(_get_path(BGPVPN_URI, fmt=self.fmt),
                                self.serialize(data),
                                content_type='application/%s' % self.fmt,
                                expect_errors=True)

            self.assertEqual(res.status_int, exc.HTTPBadRequest.code,
                             "test failed for %s" % data)

    def _test_valid_field(self, field):
        for data in self._data_for_valid_rtdt(field):
            res = self.api.post(_get_path(BGPVPN_URI, fmt=self.fmt),
                                self.serialize(data),
                                content_type='application/%s' % self.fmt,
                                expect_errors=False)

            self.assertEqual(res.status_int, exc.HTTPCreated.code,
                             "test failed for %s" % data)

    def test_bgpvpn_create_with_invalid_route_targets(self):
        self._test_invalid_field('route_targets')

    def test_bgpvpn_create_with_valid_route_targets(self):
        self._test_valid_field('route_targets')

    def test_bgpvpn_create_with_invalid_import_rts(self):
        self._test_invalid_field('import_targets')

    def test_bgpvpn_create_with_valid_import_rts(self):
        self._test_valid_field('import_targets')

    def test_bgpvpn_create_with_invalid_export_rts(self):
        self._test_invalid_field('export_targets')

    def test_bgpvpn_create_with_valid_export_rts(self):
        self._test_valid_field('export_targets')

    def test_bgpvpn_create_with_invalid_route_distinguishers(self):
        self._test_invalid_field('route_distinguishers')

    def test_bgpvpn_create_with_valid_route_distinguishers(self):
        self._test_valid_field('route_distinguishers')

    def test_bgpvpn_list(self):
        bgpvpn_id = _uuid()
        return_value = [{'name': 'bgpvpn1',
                         'type': 'l3',
                         'route_targets': ['1234:56'],
                         'id': bgpvpn_id}]

        self.instance.get_bgpvpns.return_value = return_value

        res = self.api.get(
            _get_path(BGPVPN_URI, fmt=self.fmt))

        self.instance.get_bgpvpns.assert_called_with(
            mock.ANY, fields=mock.ANY, filters=mock.ANY
        )
        self.assertEqual(res.status_int, exc.HTTPOk.code)

    def test_bgpvpn_update(self):
        bgpvpn_id = _uuid()
        update_data = {'bgpvpn': {'name': 'bgpvpn_updated'}}
        return_value = {'name': 'bgpvpn1',
                        'type': 'l3',
                        'route_targets': ['1234:56'],
                        'tenant_id': _uuid(),
                        'id': bgpvpn_id}

        self.instance.update_bgpvpn.return_value = return_value

        res = self.api.put(_get_path(BGPVPN_URI,
                                     id=bgpvpn_id,
                                     fmt=self.fmt),
                           self.serialize(update_data),
                           content_type='application/%s' % self.fmt)

        self.instance.update_bgpvpn.assert_called_with(
            mock.ANY, bgpvpn_id, bgpvpn=update_data
        )

        self.assertEqual(res.status_int, exc.HTTPOk.code)
        res = self.deserialize(res)
        self.assertIn('bgpvpn', res)
        self.assertEqual(res['bgpvpn'], return_value)

    def test_bgpvpn_get(self):
        bgpvpn_id = _uuid()
        return_value = {'name': 'bgpvpn1',
                        'type': 'l3',
                        'route_targets': ['1234:56'],
                        'tenant_id': _uuid(),
                        'id': bgpvpn_id}

        self.instance.get_bgpvpn.return_value = return_value

        res = self.api.get(_get_path(BGPVPN_URI,
                                     id=bgpvpn_id,
                                     fmt=self.fmt))

        self.instance.get_bgpvpn.assert_called_with(
            mock.ANY, bgpvpn_id, fields=mock.ANY
        )
        self.assertEqual(res.status_int, exc.HTTPOk.code)
        res = self.deserialize(res)
        self.assertIn('bgpvpn', res)
        self.assertEqual(res['bgpvpn'], return_value)

    def test_bgpvpn_delete(self):
        self._test_entity_delete('bgpvpn')

    def test_bgpvpn_net_create(self):
        data = {'network_association': {'network_id': self.net_id,
                                        'tenant_id': _uuid()}}
        return_value = copy.copy(data['network_association'])
        return_value.update({'id': self.net_assoc_id})
        self.instance.create_bgpvpn_network_association.return_value = \
            return_value
        res = self.api.post(_get_path(self.NET_ASSOC_URI, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt,
                            expect_errors=True)

        self.assertTrue(self.instance.create_bgpvpn_network_association.called)
        self.assertEqual(self.bgpvpn_id,
                         self.instance.create_bgpvpn_network_association.
                         call_args[1]['bgpvpn_id'])
        self.assertDictSupersetOf(
            data['network_association'],
            self.instance.create_bgpvpn_network_association.
            call_args[1]['network_association']['network_association'])

        self.assertIn('network_association', res)
        res = self.deserialize(res)
        self.assertDictSupersetOf(return_value,
                                  res['network_association'])

    def _invalid_data_for_creation(self, target):
        return [None, {}, {target: None}, {target: {}}]

    def test_bgpvpn_net_create_with_invalid_data(self):
        for data in self._invalid_data_for_creation('network_association'):
            res = self.api.post(_get_path(self.NET_ASSOC_URI, fmt=self.fmt),
                                self.serialize(data),
                                content_type='application/%s' % self.fmt,
                                expect_errors=True)
            self.assertFalse(
                self.instance.create_bgpvpn_network_association.called)
            self.assertEqual(res.status_int, exc.HTTPBadRequest.code)

    def test_bgpvpn_net_get(self):
        return_value = {'id': self.net_assoc_id,
                        'network_id': self.net_id}

        self.instance.get_bgpvpn_network_association.return_value = \
            return_value

        res = self.api.get(_get_path(self.NET_ASSOC_URI,
                                     id=self.net_assoc_id,
                                     fmt=self.fmt))

        self.instance.get_bgpvpn_network_association.assert_called_with(
            mock.ANY, self.net_assoc_id, self.bgpvpn_id, fields=mock.ANY
        )
        self.assertEqual(res.status_int, exc.HTTPOk.code)
        res = self.deserialize(res)
        self.assertIn('network_association', res)
        self.assertEqual(return_value, res['network_association'])

    def test_bgpvpn_net_update(self):
        pass

    def test_bgpvpn_net_delete(self):
        res = self.api.delete(_get_path(self.NET_ASSOC_URI,
                                        id=self.net_assoc_id,
                                        fmt=self.fmt))
        self.instance.delete_bgpvpn_network_association.assert_called_with(
            mock.ANY, self.net_assoc_id, self.bgpvpn_id)
        self.assertEqual(res.status_int, exc.HTTPNoContent.code)

    def test_bgpvpn_router_create(self):
        data = {
            'router_association': {
                'router_id': self.router_id,
                'tenant_id': _uuid()
            }
        }
        return_value = copy.copy(data['router_association'])
        return_value.update({'id': self.router_assoc_id})
        self.instance.create_bgpvpn_router_association.return_value = \
            return_value
        res = self.api.post(_get_path(self.ROUTER_ASSOC_URI, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt,
                            expect_errors=True)

        self.assertTrue(self.instance.create_bgpvpn_router_association.called)
        self.assertEqual(self.bgpvpn_id,
                         self.instance.create_bgpvpn_router_association.
                         call_args[1]['bgpvpn_id'])
        self.assertDictSupersetOf(
            data['router_association'],
            self.instance.create_bgpvpn_router_association.
            call_args[1]['router_association']['router_association'])

        self.assertIn('router_association', res)
        res = self.deserialize(res)
        self.assertDictSupersetOf(return_value,
                                  res['router_association'])

    def test_bgpvpn_router_create_with_invalid_data(self):
        for data in self._invalid_data_for_creation('router_association'):
            res = self.api.post(_get_path(self.ROUTER_ASSOC_URI, fmt=self.fmt),
                                self.serialize(data),
                                content_type='application/%s' % self.fmt,
                                expect_errors=True)
            self.assertFalse(
                self.instance.create_bgpvpn_router_association.called)
            self.assertEqual(res.status_int, exc.HTTPBadRequest.code)

    def test_bgpvpn_router_get(self):
        return_value = {'id': self.router_assoc_id,
                        'router_id': self.router_id}

        self.instance.get_bgpvpn_router_association.return_value = \
            return_value

        res = self.api.get(_get_path(self.ROUTER_ASSOC_URI,
                                     id=self.router_assoc_id,
                                     fmt=self.fmt))

        self.instance.get_bgpvpn_router_association.assert_called_with(
            mock.ANY, self.router_assoc_id, self.bgpvpn_id, fields=mock.ANY
        )
        self.assertEqual(res.status_int, exc.HTTPOk.code)
        res = self.deserialize(res)
        self.assertIn('router_association', res)
        self.assertEqual(return_value, res['router_association'])

    def test_bgpvpn_router_update(self):
        pass

    def test_bgpvpn_router_delete(self):
        res = self.api.delete(_get_path(self.ROUTER_ASSOC_URI,
                                        id=self.router_assoc_id,
                                        fmt=self.fmt))
        self.instance.delete_bgpvpn_router_association.assert_called_with(
            mock.ANY, self.router_assoc_id, self.bgpvpn_id)
        self.assertEqual(res.status_int, exc.HTTPNoContent.code)



# Copyright (C) 2015 Cloudwatt
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc

from networking_bgpvpn._i18n import _

# OpenContrail client API exceptions


class OpenContrailAPIFailed(n_exc.NeutronException):
    message = _("Could not reach OpenContrail API server : %(url)s "
                "Exception: %(excption)s.")


class OpenContrailAPIError(n_exc.NeutronException):
    message = _('OpenContrail API returned %(status)s %(reason)s')


class OpenContrailAPINotSupported(n_exc.BadRequest):
    message = _('OpenContrail API client cannot %(action)s on %(resource)s')


class OpenContrailAPIBadFqName(n_exc.BadRequest):
    message = _("Bad fq_name for forming a fq_name to ID request")


class OpenContrailAPIBadUUID(n_exc.BadRequest):
    message = _("Bad UUID for forming a UUID to fq_name request")


class OpenContrailAPIBadKVAttributes(n_exc.BadRequest):
    message = _("Bad attributes for forming a key/value store request")


class OpenContrailAPINotAuthorized(n_exc.NotAuthorized):
    pass


class OpenContrailAPINotFound(n_exc.NotFound):
    message = _("%(resource)s %(id)s does not exist")


class OpenContrailAPIConflict(n_exc.Conflict):
    message = _("OpenContrail API conflict: %(reason)s")


class OpenContrailAPIBadRequest(n_exc.BadRequest):
    message = _("OpenContrail API bad request: %(reason)s")


class OpenContrailMalformedUUID(n_exc.BadRequest):
    message = _("Malformed UUID: %(uuid)s")



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from django.core.urlresolvers import reverse_lazy
from django.core.validators import RegexValidator
from django.utils.translation import ugettext_lazy as _
from horizon import forms
from openstack_dashboard import api

from bgpvpn_dashboard.common import bgpvpn as bgpvpn_common
from networking_bgpvpn.neutron.services.common import constants

from bgpvpn_dashboard.dashboards.project.bgpvpn import forms \
    as project_forms


RTRD_REGEX = constants.RTRD_REGEX[4:-1]
RTRDS_REGEX = '^%s( *, *%s)*$' % (RTRD_REGEX, RTRD_REGEX)


class CommonData(project_forms.CommonData):
    route_targets = forms.CharField(
        max_length=255,
        validators=[RegexValidator(regex=RTRDS_REGEX,
                                   message=_("Route targets is not valid"))],
        label=_("Route targets"),
        required=False,
        help_text=bgpvpn_common.ROUTE_TARGET_HELP)

    import_targets = forms.CharField(
        max_length=255,
        validators=[RegexValidator(regex=RTRDS_REGEX,
                                   message=_("Import targets is not valid"))],
        label=_("Import targets"),
        required=False,
        help_text=bgpvpn_common.ROUTE_TARGET_HELP + ' To use only on import.')

    export_targets = forms.CharField(
        max_length=255,
        validators=[RegexValidator(regex=RTRDS_REGEX,
                                   message=_("Export targets is not valid"))],
        label=_("Export targets"),
        required=False,
        help_text=bgpvpn_common.ROUTE_TARGET_HELP + ' To use only on export.')

    failure_url = reverse_lazy('horizon:admin:bgpvpn:index')

    def __init__(self, request, *args, **kwargs):
        super(CommonData, self).__init__(request, *args, **kwargs)


class CreateBgpVpn(CommonData):
    tenant_id = forms.ChoiceField(label=_("Project"))

    type = forms.ChoiceField(choices=[("l3", _('l3')),
                                      ("l2", _('l2'))],
                             label=_("Type"),
                             help_text=_("The type of VPN "
                                         " and the technology behind it."))

    fields_order = ['name', 'tenant_id', 'type',
                    'route_targets', 'import_targets', 'export_targets']

    def __init__(self, request, *args, **kwargs):
        super(CreateBgpVpn, self).__init__(request, *args, **kwargs)
        tenant_choices = [('', _("Select a project"))]
        tenants, has_more = api.keystone.tenant_list(request)
        for tenant in tenants:
            if tenant.enabled:
                tenant_choices.append((tenant.id, tenant.name))
        self.fields['tenant_id'].choices = tenant_choices
        self.action = 'create'


class EditDataBgpVpn(CommonData):
    bgpvpn_id = forms.CharField(label=_("ID"), widget=forms.TextInput(
        attrs={'readonly': 'readonly'}))
    type = forms.CharField(label=_("Type"), widget=forms.TextInput(
        attrs={'readonly': 'readonly'}))
    tenant_id = forms.CharField(widget=forms.HiddenInput())

    fields_order = ['name', 'bgpvpn_id', 'tenant_id', 'type',
                    'route_targets', 'import_targets', 'export_targets']

    def __init__(self, request, *args, **kwargs):
        super(EditDataBgpVpn, self).__init__(request, *args, **kwargs)
        self.action = 'update'



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function

from neutronclient.common import extension
from neutronclient.neutron import v2_0 as neutronv20

from networking_bgpvpn._i18n import _

# To understand how neutronclient extensions work
# read neutronclient/v2.0/client.py (extend_* methods and _register_extension)


class BGPVPN(extension.NeutronClientExtension):
    resource = 'bgpvpn'
    resource_plural = '%ss' % resource

    object_path = '/bgpvpn/%s' % resource_plural
    resource_path = '/bgpvpn/%s/%%s' % resource_plural

    versions = ['2.0']


class BGPVPNCreateUpdateCommon(BGPVPN):

    def add_known_arguments(self, parser):
        """Adds to parser arguments common to create and update commands."""

        parser.add_argument(
            '--name',
            help=_('Name of the BGP VPN'))
        parser.add_argument(
            '--route-targets',
            help=_('Route Targets list to import/export for this BGP '
                   'VPN. Usage: -- --route-targets '
                   'list=true <asn1>:<nn1> <asn2>:<nn2> ...'))
        parser.add_argument(
            '--import-targets',
            help=_('List of additional Route Targets to import from.'
                   ' Usage: -- --import-targets list=true '
                   '<asn1>:<nn1> <asn2>:<nn2> ...'))
        parser.add_argument(
            '--export-targets',
            help=_('List of additional Route Targets to export to. Usage: -- '
                   '--export-targets list=true <asn1>:<nn1> <asn2>:<nn2> ...'))
        parser.add_argument(
            '--route-distinguishers',
            help=_('List of RDs that will be used to advertize VPN routes.'
                   'Usage: -- --route-distinguishers list=true '
                   '<asn1>:<nn1> <asn2>:<nn2> ...'))

    def args2body(self, parsed_args):
        body = {
            self.resource: {},
        }

        neutronv20.update_dict(parsed_args, body[self.resource],
                               ['name', 'tenant_id', 'type', 'route_targets',
                                'import_targets', 'export_targets',
                                'route_distinguishers'])

        return body


class BGPVPNCreate(BGPVPNCreateUpdateCommon,
                   extension.ClientExtensionCreate):
    """Create a BGPVPN."""
    shell_command = 'bgpvpn-create'

    def add_known_arguments(self, parser):

        BGPVPNCreateUpdateCommon.add_known_arguments(self, parser)

        # type is read-only, hence specific to create
        parser.add_argument(
            '--type',
            default='l3', choices=['l2', 'l3'],
            help=_('BGP VPN type selection between L3VPN (l3) and '
                   'EVPN (l2), default:l3'))


class BGPVPNUpdate(BGPVPNCreateUpdateCommon,
                   extension.ClientExtensionUpdate):
    """Update a given BGPVPN."""
    shell_command = 'bgpvpn-update'


class BGPVPNDelete(extension.ClientExtensionDelete,
                   BGPVPN):
    """Delete a given BGPVPN."""
    shell_command = 'bgpvpn-delete'


class BGPVPNList(extension.ClientExtensionList,
                 BGPVPN):
    """List BGPVPNs that belong to a given tenant."""
    shell_command = 'bgpvpn-list'
    list_columns = [
        'id', 'name', 'type', 'route_targets', 'import_targets',
        'export_targets', 'tenant_id', 'networks', 'routers']
    pagination_support = True
    sorting_support = True


class BGPVPNShow(extension.ClientExtensionShow,
                 BGPVPN):
    """Show a given BGPVPN."""
    shell_command = 'bgpvpn-show'


# BGPVPN  associations


def _get_bgpvpn_id(client, name_or_id):
    return neutronv20.find_resourceid_by_name_or_id(
        client, BGPVPN.resource, name_or_id)


class BGPVPNAssociation(object):

    def add_known_arguments(self, parser):
        parser.add_argument('bgpvpn', metavar='BGPVPN',
                            help=_('ID or name of the BGPVPN.'))

    def set_extra_attrs(self, parsed_args):
        self.parent_id = _get_bgpvpn_id(self.get_client(), parsed_args.bgpvpn)


# BGPVPN Network associations


class BGPVPNNetAssoc(BGPVPNAssociation,
                     extension.NeutronClientExtension):

    resource = 'network_association'
    resource_plural = '%ss' % resource

    # (parent_resource set to True so that the
    # first %s in *_path will be replaced with parent_id)
    parent_resource = True

    object_path = '%s/%s' % (BGPVPN.resource_path, resource_plural)
    resource_path = '%s/%s/%%%%s' % (BGPVPN.resource_path, resource_plural)

    versions = ['2.0']

    allow_names = False  # network associations have no name


class BGPVPNNetAssocCreate(BGPVPNNetAssoc,
                           extension.ClientExtensionCreate):
    """Create a BGPVPN-Network association."""
    shell_command = "bgpvpn-net-assoc-create"

    def add_known_arguments(self, parser):

        BGPVPNNetAssoc.add_known_arguments(self, parser)

        parser.add_argument(
            '--network', required=True,
            help=_('ID or name of the network.'))

    def args2body(self, parsed_args):

        body = {
            self.resource: {},
        }

        net = neutronv20.find_resourceid_by_name_or_id(self.get_client(),
                                                       'network',
                                                       parsed_args.network)

        body[self.resource]['network_id'] = net
        neutronv20.update_dict(parsed_args, body[self.resource], ['tenant_id'])

        return body


class BGPVPNNetAssocUpdate(extension.ClientExtensionUpdate,
                           BGPVPNNetAssoc):
    """Update a given BGPVPN-Network association."""
    shell_command = "bgpvpn-net-assoc-update"


class BGPVPNNetAssocDelete(extension.ClientExtensionDelete,
                           BGPVPNNetAssoc):
    """Delete a given BGPVPN-Network association."""
    shell_command = "bgpvpn-net-assoc-delete"


class BGPVPNNetAssocList(extension.ClientExtensionList,
                         BGPVPNNetAssoc):
    """List BGPVPN-Network associations for a given BGPVPN."""
    shell_command = "bgpvpn-net-assoc-list"

    list_columns = ['id', 'network_id']
    pagination_support = True
    sorting_support = True


class BGPVPNNetAssocShow(extension.ClientExtensionShow,
                         BGPVPNNetAssoc):
    """Show a given BGPVPN-Network association."""
    shell_command = "bgpvpn-net-assoc-show"


# BGPVPN Router associations


class BGPVPNRouterAssoc(BGPVPNAssociation,
                        extension.NeutronClientExtension):

    resource = 'router_association'
    resource_plural = '%ss' % resource
    parent_resource = True

    object_path = '%s/%s' % (BGPVPN.resource_path, resource_plural)
    resource_path = '%s/%s/%%%%s' % (BGPVPN.resource_path, resource_plural)

    versions = ['2.0']

    allow_names = False


class BGPVPNRouterAssocCreate(BGPVPNRouterAssoc,
                              extension.ClientExtensionCreate):
    """Create a BGPVPN-Router association."""
    shell_command = "bgpvpn-router-assoc-create"

    def add_known_arguments(self, parser):
        BGPVPNRouterAssoc.add_known_arguments(self, parser)
        parser.add_argument(
            '--router', required=True,
            help=_('ID or name of the router.'))

    def args2body(self, parsed_args):
        body = {
            self.resource: {},
        }
        router = neutronv20.find_resourceid_by_name_or_id(self.get_client(),
                                                          'router',
                                                          parsed_args.router)
        body[self.resource]['router_id'] = router
        neutronv20.update_dict(parsed_args, body[self.resource], ['tenant_id'])
        return body


class BGPVPNRouterAssocUpdate(extension.ClientExtensionUpdate,
                              BGPVPNRouterAssoc):
    """Update a given BGPVPN-Router association."""
    shell_command = "bgpvpn-router-assoc-update"


class BGPVPNRouterAssocDelete(extension.ClientExtensionDelete,
                              BGPVPNRouterAssoc):
    """Delete a given BGPVPN-Router association."""
    shell_command = "bgpvpn-router-assoc-delete"


class BGPVPNRouterAssocList(extension.ClientExtensionList,
                            BGPVPNRouterAssoc):
    """List BGPVPN-Router associations for a given BGPVPN."""
    shell_command = "bgpvpn-router-assoc-list"

    list_columns = ['id', 'router_id']
    pagination_support = True
    sorting_support = True


class BGPVPNRouterAssocShow(extension.ClientExtensionShow,
                            BGPVPNRouterAssoc):
    """Show a given BGPVPN-Router association."""
    shell_command = "bgpvpn-router-assoc-show"



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six

from networking_bgpvpn.neutron.db import bgpvpn_db


@six.add_metaclass(abc.ABCMeta)
class BGPVPNDriverBase(object):
    """BGPVPNDriver interface for driver

    That driver interface does not persist BGPVPN data in any database. The
    driver needs to do it by itself.
    """

    def __init__(self, service_plugin):
        self.service_plugin = service_plugin

    @property
    def service_type(self):
        pass

    @abc.abstractmethod
    def create_bgpvpn(self, context, bgpvpn):
        pass

    @abc.abstractmethod
    def get_bgpvpns(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_bgpvpn(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def update_bgpvpn(self, context, id, bgpvpn):
        pass

    @abc.abstractmethod
    def delete_bgpvpn(self, context, id):
        pass

    @abc.abstractmethod
    def create_net_assoc(self, bgpvpn_id, network_association):
        pass

    @abc.abstractmethod
    def get_net_assoc(self, context, assoc_id, bgpvpn_id, fields=None):
        pass

    @abc.abstractmethod
    def get_net_assocs(self, context, bgpvpn_id, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def delete_net_assoc(self, context, assoc_id, bgpvpn_id):
        pass

    @abc.abstractmethod
    def create_router_assoc(self, context, bgpvpn_id, router_association):
        pass

    @abc.abstractmethod
    def get_router_assoc(self, context, assoc_id, bgpvpn_id, fields=None):
        pass

    @abc.abstractmethod
    def get_router_assocs(self, context, bgpvpn_id, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def delete_router_assoc(self, context, assoc_id, bgpvpn_id):
        pass


@six.add_metaclass(abc.ABCMeta)
class BGPVPNDriverDBMixin(BGPVPNDriverBase):
    """BGPVPNDriverDB Mixin to provision the database on behalf of the driver

    That driver interface persists BGPVPN data in its database and forwards
    the result to postcommit methods
    """

    def __init__(self, service_plugin):
        super(BGPVPNDriverDBMixin, self).__init__(service_plugin)
        self.bgpvpn_db = bgpvpn_db.BGPVPNPluginDb()

    def create_bgpvpn(self, context, bgpvpn):
        with context.session.begin(subtransactions=True):
            bgpvpn = self.bgpvpn_db.create_bgpvpn(
                context, bgpvpn)
            self.create_bgpvpn_precommit(context, bgpvpn)
        self.create_bgpvpn_postcommit(context, bgpvpn)
        return bgpvpn

    def get_bgpvpns(self, context, filters=None, fields=None):
        return self.bgpvpn_db.get_bgpvpns(context, filters, fields)

    def get_bgpvpn(self, context, id, fields=None):
        return self.bgpvpn_db.get_bgpvpn(context, id, fields)

    def update_bgpvpn(self, context, id, bgpvpn):
        old_bgpvpn = self.get_bgpvpn(context, id)
        with context.session.begin(subtransactions=True):
            bgpvpn = self.bgpvpn_db.update_bgpvpn(
                context, id, bgpvpn)
            self.update_bgpvpn_precommit(context, old_bgpvpn, bgpvpn)

        self.update_bgpvpn_postcommit(context, old_bgpvpn, bgpvpn)
        return bgpvpn

    def delete_bgpvpn(self, context, id):
        with context.session.begin(subtransactions=True):
            bgpvpn = self.bgpvpn_db.delete_bgpvpn(context, id)
            self.delete_bgpvpn_precommit(context, bgpvpn)
        self.delete_bgpvpn_postcommit(context, bgpvpn)

    def create_net_assoc(self, context, bgpvpn_id, network_association):
        with context.session.begin(subtransactions=True):
            assoc = self.bgpvpn_db.create_net_assoc(context,
                                                    bgpvpn_id,
                                                    network_association)
            self.create_net_assoc_precommit(context, assoc)
        self.create_net_assoc_postcommit(context, assoc)
        return assoc

    def get_net_assoc(self, context, assoc_id, bgpvpn_id, fields=None):
        return self.bgpvpn_db.get_net_assoc(context, assoc_id, bgpvpn_id,
                                            fields)

    def get_net_assocs(self, context, bgpvpn_id, filters=None, fields=None):
        return self.bgpvpn_db.get_net_assocs(context, bgpvpn_id,
                                             filters, fields)

    def delete_net_assoc(self, context, assoc_id, bgpvpn_id):
        with context.session.begin(subtransactions=True):
            net_assoc = self.bgpvpn_db.delete_net_assoc(context,
                                                        assoc_id,
                                                        bgpvpn_id)
            self.delete_net_assoc_precommit(context, net_assoc)
        self.delete_net_assoc_postcommit(context, net_assoc)

    def create_router_assoc(self, context, bgpvpn_id, router_association):
        with context.session.begin(subtransactions=True):
            assoc = self.bgpvpn_db.create_router_assoc(context, bgpvpn_id,
                                                       router_association)
            self.create_router_assoc_precommit(context, assoc)
        self.create_router_assoc_postcommit(context, assoc)
        return assoc

    def get_router_assoc(self, context, assoc_id, bgpvpn_id, fields=None):
        return self.bgpvpn_db.get_router_assoc(context, assoc_id,
                                               bgpvpn_id, fields)

    def get_router_assocs(self, context, bgpvpn_id, filters=None, fields=None):
        return self.bgpvpn_db.get_router_assocs(context, bgpvpn_id,
                                                filters, fields)

    def delete_router_assoc(self, context, assoc_id, bgpvpn_id):
        with context.session.begin(subtransactions=True):
            router_assoc = self.bgpvpn_db.delete_router_assoc(context,
                                                              assoc_id,
                                                              bgpvpn_id)
            self.delete_router_assoc_precommit(context, router_assoc)
        self.delete_router_assoc_postcommit(context, router_assoc)

    @abc.abstractmethod
    def create_bgpvpn_postcommit(self, context, bgpvpn):
        pass

    @abc.abstractmethod
    def create_bgpvpn_precommit(self, context, bgpvpn):
        pass

    @abc.abstractmethod
    def update_bgpvpn_postcommit(self, context, old_bgpvpn, bgpvpn):
        pass

    @abc.abstractmethod
    def update_bgpvpn_precommit(self, context, old_bgpvpn, bgpvpn):
        pass

    @abc.abstractmethod
    def delete_bgpvpn_postcommit(self, context, bgpvpn):
        pass

    @abc.abstractmethod
    def create_net_assoc_precommit(self, context, net_assoc):
        pass

    @abc.abstractmethod
    def create_net_assoc_postcommit(self, context, net_assoc):
        pass

    @abc.abstractmethod
    def delete_net_assoc_precommit(self, context, net_assoc):
        pass

    @abc.abstractmethod
    def delete_net_assoc_postcommit(self, context, net_assoc):
        pass

    @abc.abstractmethod
    def create_router_assoc_precommit(self, context, router_assoc):
        pass

    @abc.abstractmethod
    def create_router_assoc_postcommit(self, context, router_assoc):
        pass

    @abc.abstractmethod
    def delete_router_assoc_precommit(self, context, router_assoc):
        pass

    @abc.abstractmethod
    def delete_router_assoc_postcommit(self, context, router_assoc):
        pass


class BGPVPNDriver(BGPVPNDriverDBMixin):
    """BGPVPNDriver interface for driver with database.

    Each bgpvpn driver that needs a database persistency should inherit
    from this driver.
    It can overload needed methods from the following pre/postcommit methods.
    Any exception raised during a precommit method will result in not having
    related records in the databases.
    """

    def create_bgpvpn_precommit(self, context, bgpvpn):
        pass

    def create_bgpvpn_postcommit(self, context, bgpvpn):
        pass

    def update_bgpvpn_precommit(self, context, old_bgpvpn, bgpvpn):
        pass

    def update_bgpvpn_postcommit(self, context, old_bgpvpn, bgpvpn):
        pass

    def delete_bgpvpn_precommit(self, context, bgpvpn):
        pass

    def delete_bgpvpn_postcommit(self, context, bgpvpn):
        pass

    def create_net_assoc_precommit(self, context, net_assoc):
        pass

    def create_net_assoc_postcommit(self, context, net_assoc):
        pass

    def delete_net_assoc_precommit(self, context, net_assoc):
        pass

    def delete_net_assoc_postcommit(self, context, net_assoc):
        pass

    def create_router_assoc_precommit(self, context, router_assoc):
        pass

    def create_router_assoc_postcommit(self, context, router_assoc):
        pass

    def delete_router_assoc_precommit(self, context, router_assoc):
        pass

    def delete_router_assoc_postcommit(self, context, router_assoc):
        pass
